@@ -217,12 +217,11 @@
 (declaim (inline address-based-counter-val quasi-random-address-based-hash))
 (defun address-based-counter-val ()
   (let ((word
-         ;; threads imply gencgc. use the per-thread alloc region pointer
-         #+sb-thread
+         ;; Use the per-thread alloc region pointer when possible
+         #+(or x86-64 sb-thread)
          (sap-int (sb-vm::current-thread-offset-sap sb-vm::thread-mixed-tlab-slot))
-         ;; dynamic-space-free-pointer increments only when a page is full.
-         ;; Using mixed_region directly is finer-grained.
-         #+(and (not sb-thread) gencgc)
+         ;; Otherwise mixed_region in static space
+         #-(or x86-64 sb-thread)
          (sb-sys:sap-ref-word (sb-sys:int-sap (+ sb-vm::static-space-start
                                                  sb-vm::mixed-region-offset))
                               0)))
@@ -259,15 +258,15 @@
 ;;; You could always make it SB-XC:FIXNUM at the risk of forcing the host to
 ;;; deal in bignums. Why cause it undue slowness when we don't need so many bits?
 ;;; NOTE: we _do_ use the sign bit, leaving us 25 pseudorandom bits, but
-;;; the 2 bits of least significance are NOT pseudorandom, so it's best
+;;; the 3 bits of least significance are NOT pseudorandom, so it's best
 ;;; not to use them directly in the hash index.
-(eval-when (:compile-toplevel :load-toplevel :execute)
-  (defconstant ctype-hash-size  30)  ; all significant bits, for the slot type specifier
-  (defconstant ctype-PRNG-nbits 25)  ; from pseudorandom number generator
-  (defconstant ctype-contains-unknown #b01)
-  (defconstant ctype-contains-hairy   #b10)) ; any hairy type, including UNKNOWN
-(defconstant +ctype-flag-mask+ #b11)
-(defconstant +ctype-hash-mask+ (logandc2 (1- (ash 1 ctype-PRNG-nbits)) #b11))
+(defconstant ctype-hash-size  30)  ; all significant bits, for the slot type specifier
+(defconstant ctype-PRNG-nbits 25)  ; from pseudorandom number generator
+(defconstant ctype-contains-unknown #b001)
+(defconstant ctype-contains-hairy   #b010) ; any hairy type, including UNKNOWN
+(defconstant ctype-contains-class   #b100) ; standard-class
+(defconstant +ctype-flag-mask+ #b111)
+(defconstant +ctype-hash-mask+ (logandc2 (1- (ash 1 ctype-PRNG-nbits)) #b111))
 
 (defstruct (ctype (:conc-name type-)
                    (:constructor nil)
@@ -279,17 +278,23 @@
 
 ;;; Apparently the old CONTAINS-UNKNOWN-TYPE-P function could accept NIL
 ;;; and return NIL. This seems kinda sloppy. Can we get rid of that "feature"?
-(declaim (inline contains-unknown-type-p contains-hairy-type-p))
+(declaim (inline contains-unknown-type-p contains-hairy-type-p opaque-type-p))
 (defun contains-unknown-type-p (ctype)
-  (if ctype (oddp (type-%bits ctype)) nil))
+  (and ctype
+       (logtest (type-%bits ctype) ctype-contains-unknown)))
 (defun contains-hairy-type-p (ctype)
-  (logbitp 1 (type-%bits ctype)))
+  (logtest (type-%bits ctype) ctype-contains-hairy))
+
+;;; Can't do optimizations for satisfies, unknown types or standard-class.
+(defun opaque-type-p (ctype)
+  (logtest (type-%bits ctype) +ctype-flag-mask+))
 
 (defun ok-to-memoize-p (arg)
   (etypecase arg
-    (ctype (evenp (type-%bits arg))) ; i.e. not CTYPE-CONTAINS-UNKNOWN
+    (ctype (not (logtest (type-%bits arg) ctype-contains-unknown)))
     (list  (dolist (elt arg t)
-             (when (oddp (type-%bits elt)) (return nil))))))
+             (when (logtest (type-%bits elt) ctype-contains-unknown)
+               (return nil))))))
 
 (defmacro type-class-id (ctype) `(ldb (byte 5 ,ctype-PRNG-nbits) (type-%bits ,ctype)))
 (defmacro type-id->type-class (id) `(truly-the type-class (aref *type-classes* ,id)))
@@ -333,7 +338,7 @@
       (intersection  intersection-type)
       (union         union-type)
       (negation      negation-type)
-      (number        numeric-type)
+      (numeric-union numeric-union-type)
       (array         array-type)
       (character-set character-set-type)
       (member        member-type)
@@ -652,7 +657,7 @@
 ;;;  2. it inserts (:COPIER NIL)
 ;;;  3. it adds :READ-ONLY T to all slots
 ;;;  4. it has slot options to help with hash-consing
-(defmacro def-type-model ((name &rest options) &rest direct-slots)
+(defmacro def-type-model ((name &rest options) &body direct-slots)
   ;; :CONSTRUCTOR* reminds you that it's not a direct translation to defstruct.
   (aver (<= (count :constructor* options :key #'car) 1))
   ;; The private constructor is always positional.
@@ -673,7 +678,7 @@
          ;; Inherited slots are mixed in by calling the supertype's hash function.
          (hashed-slots
           (mapcan (lambda (slot &aux (type (getf (cddr slot) :type))
-                                     (hasher (getf (cddr slot) :hasher 'clipped-sxhash))
+                                     (hasher (getf (cddr slot) :hasher :fail))
                                      (comparator (getf (cddr slot) :test)))
                     (cond ((or comparator (null hasher))) ; ok
                           ((eq type 'ctype)
@@ -681,7 +686,9 @@
                           ((or (and (typep type '(cons (eql member)))
                                     (every #'symbolp (cdr type)))
                                (eq type 'boolean))
-                           (setq comparator 'eq))
+                           (setq comparator 'eq)
+                           (when (eq type 'boolean)
+                             (setq hasher '(lambda (x) (if x #xaa55aa #x55aa55)))))
                           (t (bug "Underspecified slot: ~S.~S type ~S~%"
                                   name (car slot) type)))
                     (when hasher (list (vector (car slot) hasher comparator))))
@@ -703,9 +710,9 @@
                    direct-slots))
        ;; Always define hash-consing functions if any new slots, even if abstract.
        ,@(when direct-slots
-           (let ((hashfn (symbolicate "CALC-" name "-HASH"))
-                 (compare (symbolicate name "-EQUIV")))
-             `((defun ,hashfn (x)
+           (let
+            ((functions
+             `((defun ,(symbolicate "CALC-" name "-HASH") (x)
                  #-sb-xc-host (declare (optimize (safety 0)))
                  (,(if (assoc :extra-mix-step options) 'type-hash-final-mix 'progn)
                   (type-hash-mix
@@ -714,13 +721,14 @@
                                                  (symbolicate conc-name (svref slot 0))))
                                `(,(svref slot 1) (,reader x)))
                              hashed-slots))))
-               (defun ,compare (a b)
+               (defun ,(symbolicate name "-EQUIV") (a b)
                  #-sb-xc-host (declare (optimize (safety 0)))
                  (and ,@(mapcar (lambda (slot &aux (reader
                                                     (symbolicate conc-name (svref slot 0))))
                                   `(,(elt slot 2) (,reader a) (,reader b)))
                                 hashed-slots)
                       ,@(when include `((,(symbolicate include "-EQUIV") a b))))))))
+             functions))
        ;; Define a hashset unless this is an abstract type
        ,@(unless (member name '(compound-type args-type))
            (let* ((stem (if direct-slots name include))
@@ -745,31 +753,21 @@
                              (array-type '(type-flags element-type)))
                           ,@(cdr private-ctor-args))))))))
 
-;;; The "clipped hash" is just some stable hash that may rely on the host's SXHASH
-;;; but always ensuring that the result is an unsigned fixnum for the target,
-;;; so that we can call our MIX on the value. It's needlessly tedious to fully
-;;; replicate our SXHASH on BIGNUM and RATIO which can appear in numeric bounds.
-#+sb-xc-host
-(defun clipped-sxhash (x)
-  (typecase x
-    (rational ; numeric-type high,low bound; array dimensions, etc
-     ;; All integers can fall through to the host because it would be silly to restrict
-     ;; this case to exactly (OR (AND INTEGER (NOT SB-XC:FIXNUM)) RATIO).
-     (logand (cl:sxhash x) sb-xc:most-positive-fixnum))
-    (cons
-     (if (eq (car x) 'satisfies)
-         (sb-xc:sxhash (cadr x)) ; it's good enough
-         (error "please no: ~S" x)))
-    (t
-     (sb-xc:sxhash x)))) ; FLOAT representation as struct, or SYMBOL
-#-sb-xc-host (defmacro clipped-sxhash (x) `(sxhash ,x))
-
 (defmacro type-hash-mix (&rest args) (reduce (lambda (a b) `(mix ,a ,b)) args))
 ;;; The final mix ensures that all bits affect the masked hash.
 ;;; Since it takes non-zero time, only do it for NUMERIC and ARRAY, where it seems
 ;;; to make a large difference in the maximum probe sequence length.
 (defmacro type-hash-final-mix (val) `(murmur-hash-word/+fixnum ,val))
 
+#-sb-xc-host
+(progn
+  (defmacro sb-c::number-hash (x) `(sb-impl::number-sxhash ,x))
+  ;; This is used on a HAIRY specifier which could be an UNKNOWN (just a symbol), or a SATISFIES.
+  ;; There is no reason at all that two distinct symbols should hash the same when their
+  ;; names are STRING= so really we want something better than SXHASH, but it does noeed to
+  ;; recurse on lists.
+  (defmacro sb-c::fallback-hash (x) `(sxhash ,x)))
+
 ;; Singleton MEMBER types are best dealt with via a weak-value hash-table because:
 ;; * (MEMBER THING) might lack an address-insensitive hash for THING
 ;;   but src/code/hashset goes through a lot of rigmarole to handle address-bashed
@@ -789,10 +787,6 @@
 (define-load-time-global *eql-type-cache* ; like EQL-SPECIALIZER-TABLE in PCL
     (sb-impl::make-system-hash-table :test 'eql :weakness :value :synchronized nil))
 
-(defmacro safe-member-type-elt-p (obj)
-  `(or (not (sb-vm:is-lisp-pointer (get-lisp-obj-address ,obj)))
-       (heap-allocated-p ,obj)))
-
 #-sb-xc-host
 (defun ctype-hashset-insert-if-absent (hashset key function)
   (or (hashset-find hashset key)
@@ -833,7 +827,7 @@
     ;; COPY-CTYPE can copy subparts like the numeric bound if arena-allocated
     #-sb-xc-host
     `(let ((temp (,allocator (logior ,flag-bits ,class-bits) ,@initargs)))
-       (declare (truly-dynamic-extent temp))
+       (declare (dynamic-extent temp))
        #+nil ; or #+sb-devel as you see fit
        (unless *hashsets-preloaded*
          (write-string "CTYPE hashset preload failure")
@@ -876,21 +870,13 @@
   ;; is that apparently we'll store _illegal_ type specifiers in a hairy-type.
   ;; There's an example in the regression test named
   ;;  :single-warning-for-single-undefined-type
-  (specifier nil :type t :test equal))
+  (specifier nil :type t :test equal :hasher sb-c::fallback-hash))
 
-(macrolet ((hash-fp-zeros (x) ; order-insensitive
-             `(let ((h 0))
-                (dolist (x ,x h) (setq h (logxor (sb-xc:sxhash x) h)))))
-           (fp-zeros= (a b)
-             `(let ((a ,a) (b ,b))
-                (and (= (length a) (length b))
-                     (every (lambda (x) (member x b)) a)))))
 ;;; A MEMBER-TYPE represent a use of the MEMBER type specifier. We
 ;;; bother with this at this level because MEMBER types are fairly
 ;;; important and union and intersection are well defined.
-(def-type-model (member-type (:constructor* nil (xset fp-zeroes)))
-  (xset nil :type xset :hasher xset-elts-hash :test xset=)
-  (fp-zeroes nil :type list :hasher hash-fp-zeros :test fp-zeros=)))
+(def-type-model (member-type (:constructor* nil (xset)))
+  (xset nil :type xset :hasher xset-elts-hash :test xset=))
 (define-load-time-global *xset-mutex* (or #-sb-xc-host (sb-thread:make-mutex :name "xset")))
 ;;; This hashset is guarded by *XSET-MUTEX*. It is _not_ declared as synchronized
 ;;; so that HASHSET-INSERT-IF-ABSENT should not acquire a mutex inside a mutex
@@ -923,7 +909,8 @@
   ;; is unspecified, it is *.
   (dimensions '* :type (or list (eql *)) :test equal :hasher hash-dims)
   ;; Is this not a simple array type? (:MAYBE means that we don't know.)
-  (complexp :maybe :type (member t nil :maybe))
+  (complexp :maybe :type (member t nil :maybe)
+            :hasher (lambda (s) (case s (:maybe #36rMAYBE) ((nil) #xffff) (t 1))))
   ;; the element type as originally specified
   (element-type nil :type ctype)
   ;; the element type as it is specialized in this implementation
@@ -1073,23 +1060,19 @@
              (setf (aref *numeric-aspects-v* index)
                    (!make-numeric-aspects complexp class precision index)))))
 
-(defmacro get-numtype-aspects (&rest rest)
+(defmacro get-numtype-aspects (complexp class precision)
   `(the (not null)
-        (aref *numeric-aspects-v* (!compute-numtype-aspect-id ,@rest))))
+        (aref *numeric-aspects-v* (!compute-numtype-aspect-id ,complexp ,class ,precision))))
 
 (macrolet ((numbound-hash (b)
              ;; It doesn't matter what the hash of a number is, as long as it's stable.
-             ;; Use the host's SXHASH for convenience.
-             ;; We aren't obliged to fully emulate own behavior on numbers,
-             ;; but we can't trust the host to do the right thing on our proxy floats.
              `(let ((x ,b))
                 (block nil
                   (multiple-value-bind (h v)
                       (if (listp x)
                           (if x (values #x55AA55 (car x)) (return 0))
                           (values 0 x))
-                    (logxor h (#+sb-xc-host clipped-sxhash
-                               #-sb-xc-host sb-impl::number-sxhash v))))))
+                    (logxor h (sb-c::number-hash v))))))
            (numbound-eql (a b)
              ;; Determine whether the 'low' and 'high' slots of two NUMERIC-TYPE instances
              ;; are "the same". It is a stricter test than in the SIMPLE-= method, because
@@ -1098,22 +1081,65 @@
              `(let ((a ,a) (b ,b))
                 (if (listp a)
                     (and (listp b) (eql (car a) (car b)))
-                    (eql a b)))))
-;;; A NUMERIC-TYPE represents any numeric type, including things
-;;; such as FIXNUM.
-(def-type-model (numeric-type
-                 (:extra-mix-step)
-                 (:constructor* nil (aspects low high)))
-  (aspects (missing-arg) :type numtype-aspects :hasher numtype-aspects-id :test eq)
-  (low nil :type (or real (cons real null) null)
-       :hasher numbound-hash :test numbound-eql)
-  (high nil :type (or real (cons real null) null)
-        :hasher numbound-hash :test numbound-eql)))
+                    (eql a b))))
+           (hash-ranges (a)
+             `(let ((vector ,a)
+                    (h 0))
+                (loop for e across vector
+                      do
+                      (setf h (mix h (numbound-hash e))))
+                h)))
+
+  (def-type-model (numeric-union-type
+                   (:extra-mix-step)
+                   (:constructor* nil (aspects ranges)))
+    (aspects (missing-arg) :type numtype-aspects :hasher numtype-aspects-id :test eq)
+    ;; Ranges are sorted in ascending order by their low bound.
+    ;; Rational ranges are represented by three entries,
+    ;; #(run low high ...) where run is one of range-integer-run,
+    ;; range-ratio-run, range-rational-run.
+    ;; Floats are just #(low high ...)
+    (ranges #() :type simple-vector :hasher hash-ranges :test equalp)))
+
+(declaim (inline numeric-type-aspects))
+(defun numeric-type-aspects (x)
+  (numeric-union-type-aspects x))
+
 (declaim (inline numeric-type-complexp numeric-type-class numeric-type-format))
 (defun numeric-type-complexp (x) (numtype-aspects-complexp (numeric-type-aspects x)))
 (defun numeric-type-class (x) (numtype-aspects-class (numeric-type-aspects x)))
 (defun numeric-type-format (x) (numtype-aspects-precision (numeric-type-aspects x)))
 
+;;; A single-range type. Similar to the old model.
+(deftype numeric-type () `(satisfies numeric-type-p))
+
+(defun numeric-type-p (x)
+  (typecase x
+    (numeric-union-type
+     (<= (length (numeric-union-type-ranges x)) 3))))
+
+(defun numeric-type-low (x)
+  (let ((ranges (numeric-union-type-ranges x)))
+    (ecase (length ranges)
+      (3 (aref ranges 1))
+      (2 (aref ranges 0)))))
+
+(defun numeric-type-high (x)
+  (let ((ranges (numeric-union-type-ranges x)))
+    (ecase (length ranges)
+      (3 (aref ranges 2))
+      (2 (aref ranges 1)))))
+
+(defun numeric-union-type-low (x)
+  (let ((ranges (numeric-union-type-ranges x)))
+    (case (numeric-type-class x)
+      ((integer rational) (aref ranges 1))
+      (t (aref ranges 0)))))
+
+(defun numeric-union-type-high (x)
+  (let ((ranges (numeric-union-type-ranges x)))
+    (aref ranges (1- (length ranges)))))
+
 ;;; A CONS-TYPE is used to represent a CONS type.
 (def-type-model (cons-type (:constructor* nil (car-type cdr-type)))
   ;; the CAR and CDR element types (to support ANSI (CONS FOO BAR) types)
@@ -1158,11 +1184,8 @@
        (eq (key-info-type a) (key-info-type b))))
 (defun key-info-hash (x)
   (declare (optimize (safety 0)))
-  (#+sb-xc-host progn
-   #-sb-xc-host murmur-hash-word/fixnum ; scramble some more
-   (mix (#+sb-xc-host sb-impl::symbol-name-hash #-sb-xc-host sxhash (key-info-name x))
-        ;; MIX requires positive fixnums
-        (logand (type-hash-value (key-info-type x)) sb-xc:most-positive-fixnum))))
+  (murmur-hash-word/+fixnum
+   (mix (type-hash-value (key-info-type x)) (symbol-hash (key-info-name x)))))
 
 (defun hash-key-info-set (set) ; order-insensitive
   (declare (optimize (safety 0)))
@@ -1171,7 +1194,7 @@
     ;; Don't need the answer to be positive for key-info-set-hashset,
     ;; but do need it to be positive when hashing ARGS-TYPE which uses MIX.
     (dolist (elt set (logand h sb-xc:most-positive-fixnum))
-      (setf h (plus-mod-fixnum (truly-the fixnum (key-info-hash elt)) h)))))
+      (setf h (plus-mod-fixnum (truly-the sb-xc:fixnum (key-info-hash elt)) h)))))
 
 (defun key-info-list-flags (list)
   (let ((bits 0))
@@ -1239,7 +1262,7 @@
 (def-type-model (simd-pack-type
                  (:constructor* %make-simd-pack-type (tag-mask)))
   (tag-mask (missing-arg) ; bitmask over possible simd-pack-tag values
-   :test =
+   :test = :hasher identity ; the tag-mask is its own hash
    :type (and (unsigned-byte #.(length +simd-pack-element-types+))
               (not (eql 0)))))
 
@@ -1247,7 +1270,7 @@
 (def-type-model (simd-pack-256-type
                  (:constructor* %make-simd-pack-256-type (tag-mask)))
   (tag-mask (missing-arg)
-   :test =
+   :test = :hasher identity ; the tag-mask is its own hash
    :type (and (unsigned-byte #.(length +simd-pack-element-types+))
               (not (eql 0)))))
 
@@ -1343,64 +1366,6 @@
               caches)))
       (display caches))))
 
-#-sb-xc-host
-(progn
-(defglobal *!initial-ctypes* nil)
-(defun preload-ctype-hashsets ()
-  (dolist (pair (nreverse *!initial-ctypes*))
-    (let ((instance (car pair))
-          (container (symbol-value (cdr pair))))
-      (cond ((hash-table-p container)
-             (aver (member-type-p instance))
-             ;; As of this writing there are only two EQL types to preload:
-             ;; one is in the IR1-transform of FORMAT with stream (EQL T),
-             ;; the other is CHECK-ARG-TYPE looking for (EQL DUMMY) type.
-             (let ((key (first (member-type-members instance))))
-               (aver (not (gethash key container)))
-               (setf (gethash key container) instance)))
-            (t
-             (aver (not (hashset-find container instance))) ; instances are built bottom-up
-             (hashset-insert container instance)))
-      (labels ((ensure-interned-list (list hashset)
-                 (let ((found (hashset-find hashset list)))
-                   (when (and found (neq found list))
-                     (bug "genesis failed to uniquify list-of-ctype in ~X"
-                          (get-lisp-obj-address instance)))
-                   (when (and list (not found))
-                     (hashset-insert hashset list)))
-                 (mapc #'check list))
-               ;; Assert that looking for SUBPART finds nothing or finds itself
-               (check (subpart &aux (hashset-symbol (ctype->hashset-sym subpart)))
-                 (when hashset-symbol
-                   (let* ((hashset (symbol-value hashset-symbol))
-                          (found (hashset-find hashset subpart)))
-                     (when (and found (neq found subpart))
-                       (bug "genesis dumped bad instance within ~X"
-                            (get-lisp-obj-address instance)))))))
-        (etypecase instance
-          ((or numeric-type member-type character-set-type ; nothing extra to do
-           #+sb-simd-pack simd-pack-type #+sb-simd-pack-256 simd-pack-256-type
-           hairy-type))
-          (args-type
-           (ensure-interned-list (args-type-required instance) *ctype-list-hashset*)
-           (ensure-interned-list (args-type-optional instance) *ctype-list-hashset*)
-           (awhen (args-type-rest instance) (check it))
-           (when (fun-type-p instance)
-             (aver (null (fun-type-keywords instance)))
-             (check (fun-type-returns instance))))
-          (cons-type
-           (check (cons-type-car-type instance))
-           (check (cons-type-cdr-type instance)))
-          (array-type
-           (check (array-type-element-type instance))
-           (check (array-type-specialized-element-type instance)))
-          (compound-type
-           (ensure-interned-list (compound-type-types instance) *ctype-set-hashset*))
-          (negation-type
-           (check (negation-type-type instance)))))))
-  #+sb-devel (setq *hashsets-preloaded* t))
-(preload-ctype-hashsets))
-
 ;;; *TYPE-CLASS-LIST* is defined only in the host. When the cross-compiler
 ;;; expands TYPEP-IMPL-MACRO it get the value of this symbol from the host's
 ;;; value. This function avoids a warning about a missing symbol.
@@ -1538,101 +1503,6 @@
                (and (not (complexp object))
                     (bound-test object))))))))
 
-;;; Copy X to the heap, give it a random hash, and if it is a MEMBER type
-;;; then assert that all members are cacheable.
-#+sb-xc-host
-(defun copy-ctype (x)
-  (let ((bits (logior (type-%bits x) (logand (ctype-random) +ctype-hash-mask+))))
-    (etypecase x
-      (member-type
-       (!alloc-member-type bits (member-type-xset x) (member-type-fp-zeroes x))))))
-#-sb-xc-host
-(defun copy-ctype (x &optional (flags 0))
-  (declare (type ctype x))
-  (declare (sb-c::tlab :system) (inline !new-xset))
-  #+c-stack-is-control-stack (aver (stack-allocated-p x))
-  (labels ((copy (x)
-             ;; Return a heap copy of X if X was arena or stack-allocated.
-             ;; I suspect it's quicker to copy always rather than conditionally.
-             ;; The use for this is that supposing the user constructs a type specifier
-             ;; like (DOUBLE-FLOAT (2.0) 4.0) where those numbers and the inner list
-             ;; were constructed on an arena, they need to be copied.
-             (etypecase x
-               (number (sb-vm:copy-number-to-heap x))
-               (cons (cons (copy (car x)) (copy (cdr x))))
-               (symbol x)))
-           (copy-xset (xset &aux (data (xset-data xset)))
-             ;; MEMBER-TYPE is a problem because the members could be arena-allocated.
-             ;; It would be easy enough to avoid entering some instances in a hashset, though
-             ;; the larger issue is that it may be inserted into any number of other caches.
-             ;; CLHS never says whether DX objects are or aren't legal in type specifiers.
-             ;; I consider this "user error" as it seems to push the boundary of what should
-             ;; be permissible, but we can do better than to cache data that are on the stack.
-             ;; If the XSET is represented as a hash-table, we may have another issue
-             ;; which is not dealt with here (hash-table in the arena)
-             (cond ((listp data)
-                    ;; the XSET can be empty if a MEMBER type contains only FP zeros.
-                    ;; While we could use (load-time-value) to referece a constant empty xset
-                    ;; there's really no point to doing that.
-                    (collect ((elts))
-                      (dolist (x data (!new-xset (elts) (xset-extra xset)))
-                        (elts (cond ((numberp x) (sb-vm:copy-number-to-heap x))
-                                    ((safe-member-type-elt-p x) x)
-                                    ;; surely things will go haywire if this occurs
-                                    (t (error "Off-heap MEMBER type member @ ~X"
-                                              (get-lisp-obj-address x))))))))
-                   ;; Huge MEMBER types are rare so I'm not going to worry too much,
-                   ;; just check whether it's OK or not
-                   ((and (loop for k being each hash-key of data
-                               always (safe-member-type-elt-p k))
-                         (heap-allocated-p data))
-                    xset)
-                   (t ; This could certainly be improved
-                    (error "Off-heap MEMBER type members")))))
-    (let ((bits (logior (type-%bits x) (logand (ctype-random) +ctype-hash-mask+) flags)))
-      ;; These cases are in descending order of frequency of seek in the hashsets.
-      ;; It matters only for backends that don't convert TYPECASE to a jump table.
-      (etypecase x
-        (values-type
-         (!alloc-values-type bits (values-type-required x) (values-type-optional x)
-                             (values-type-rest x)))
-        (fun-type ; or FUN-DESIGNATOR-TYPE
-         (let ((copy (!alloc-fun-type
-                      bits (fun-type-required x) (fun-type-optional x) (fun-type-rest x)
-                      (fun-type-keyp x) (fun-type-keywords x) (fun-type-allowp x)
-                      (fun-type-wild-args x) (fun-type-returns x))))
-           (%set-instance-layout copy (%instance-layout x))
-           copy))
-        (numeric-type
-         (!alloc-numeric-type bits (numeric-type-aspects x)
-                              (copy (numeric-type-low x)) (copy (numeric-type-high x))))
-        (compound-type ; UNION or INTERSECTION
-         (let ((copy (!alloc-union-type bits (compound-type-enumerable x)
-                                        (compound-type-types x))))
-           (%set-instance-layout copy (%instance-layout x))
-           copy))
-        (member-type
-         (!alloc-member-type bits (copy-xset (member-type-xset x))
-          (mapcar 'sb-vm:copy-number-to-heap (member-type-fp-zeroes x))))
-        (array-type
-         (!alloc-array-type bits (copy (array-type-dimensions x))
-                            (array-type-complexp x) (array-type-element-type x)
-                            (array-type-specialized-element-type x)))
-        (hairy-type ; SATISFIES or UNKNOWN
-         (let ((copy (!alloc-hairy-type bits (copy (hairy-type-specifier x)))))
-           (%set-instance-layout copy (%instance-layout x))
-           copy))
-        (negation-type (!alloc-negation-type bits (negation-type-type x)))
-        (constant-type (!alloc-constant-type bits (constant-type-type x)))
-        (cons-type (!alloc-cons-type bits (cons-type-car-type x) (cons-type-cdr-type x)))
-        (character-set-type
-         (!alloc-character-set-type bits (copy (character-set-type-pairs x))))
-        #+sb-simd-pack
-        (simd-pack-type (!alloc-simd-pack-type bits (simd-pack-type-tag-mask x)))
-        #+sb-simd-pack-256
-        (simd-pack-256-type (!alloc-simd-pack-256-type bits (simd-pack-256-type-tag-mask x)))
-        (alien-type-type (!alloc-alien-type-type bits (alien-type-type-alien-type x)))))))
-
 ;;; Drop NILs, possibly reducing the storage vector length
 (defun rebuild-ctype-hashsets ()
   (dolist (sym (list* '*key-info-hashset* '*key-info-set-hashset*
