@@ -16,21 +16,29 @@
 ;;;; files for more information.
 
 (defpackage "SB-COLD"
-  (:use "CL"))
+  (:use "CL")
+  (:export
+   "BACKEND-SUBFEATURES"
+   "FIND-BOOTSTRAP-FILE"
+   "PRESERVING-HOST-FUNCTION"
+   "STEM-SOURCE-PATH"))
 (in-package "SB-COLD")
 
-(setq *features* (union *features* sb-impl:+internal-features+))
+(setq *features* (cons :sb-xc (union *features* sb-impl:+internal-features+)))
 
-(defstruct package-data name doc shadow export reexport import-from use)
-(export 'package-data)
-(dolist (data (with-open-file (f (merge-pathnames "../../package-data-list.lisp-expr"
-                                                  *load-pathname*))
-                (read f)))
-  (labels ((flatten (tree)
-             (mapcan (lambda (x) (if (listp x) (flatten x) (list x)))
-                     tree)))
-    (let ((*package* (find-package (package-data-name data))))
-      (export (mapcar 'intern (flatten (package-data-export data)))))))
+(defun backend-asm-package-name ()
+  (package-name sb-assem::*backend-instruction-set-package*))
+
+(defun target-platform-keyword ()
+  (let ((arch (intersection '(:arm :arm64 :mips :ppc :ppc64 :riscv :sparc :x86 :x86-64) *features*)))
+    (assert (= (length arch) 1))
+    (car arch)))
+
+(defun backend-assembler-target-name ()
+  (let ((keyword (target-platform-keyword)))
+    (case keyword
+      (:ppc :ppc64)
+      (t keyword))))
 
 (sb-ext:unlock-package "CL")
 (rename-package "COMMON-LISP" "COMMON-LISP"
@@ -41,16 +49,11 @@
     (when (sb-int:system-package-p (find-package name))
       (sb-ext:unlock-package package))))
 
-;;; Define this first to avoid a style-warning from 'shebang'
-(defun read-target-float (stream char)
-  (declare (ignore stream char))
-  (values)) ; ignore the $ as if it weren't there
-(compile 'read-target-float)
-(set-macro-character #\$ #'read-target-float t)
+(handler-bind ((sb-int:package-at-variance #'muffle-warning))
+  (load (merge-pathnames "exports.lisp" *load-pathname*)))
 
-;; Restore !DEFINE-LOAD-TIME-GLOBAL macro
-(setf (macro-function 'sb-int::!define-load-time-global)
-      (macro-function 'sb-ext:define-load-time-global))
+(rename-package "SB-SEQUENCE" "SB-SEQUENCE"
+                (cons "SEQUENCE" (package-nicknames "SB-SEQUENCE")))
 
 (unless (fboundp 'sb-int:!cold-init-forms)
   (defmacro sb-int:!cold-init-forms (&rest forms) `(progn ,@forms)))
@@ -63,8 +66,6 @@
         (macro-function 'sb-int:/show0) (macro-function 'sb-int:/show)
         (macro-function 'sb-int:/noshow0) (macro-function 'sb-int:/show)))
 
-sb-c::(defun %vop-existsp (name query &optional optimistic)
-  (declare (ignore optimistic))
-  (not (null (ecase query
-               (:translate (awhen (info :function :info name) (fun-info-templates it)))
-               (:named (gethash name *backend-template-names*))))))
+(defmacro preserving-host-function (form) form)
+
+(load "SYS:src;compiler;vop-existsp.lisp")
